use std::fmt::{self, Write};

#[cfg(feature = "serde")]
use serde::{Deserialize, Serialize};

#[derive(Clone, Debug, PartialEq)]
#[cfg_attr(feature = "serde", derive(Serialize, Deserialize))]
pub enum Node {
    Text(String),
    H1(String),
    H2(String),
    H3(String),
    ListItem(String),
    Quote(String),
    PreStart {
        alt: Option<String>,
    },
    PreBody(String),
    PreEnd,
    Link {
        url: String,
        display: Option<String>,
    },
    Prompt {
        url: String,
        display: Option<String>,
    },
}

impl fmt::Display for Node {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            Self::Text(s) | Self::PreBody(s) => writeln!(f, "{s}"),
            Self::H1(s) => writeln!(f, "# {s}"),
            Self::H2(s) => writeln!(f, "## {s}"),
            Self::H3(s) => writeln!(f, "### {s}"),
            Self::ListItem(s) => writeln!(f, "* {s}"),
            Self::Quote(s) => writeln!(f, "> {s}"),
            Self::PreStart { alt } => match alt {
                Some(s) => writeln!(f, "```{s}"),
                None => writeln!(f, "```"),
            },
            Self::PreEnd => writeln!(f, "```"),
            Self::Link {
                url,
                display: Some(display),
            } => writeln!(f, "=> {url} {display}"),
            Self::Link { url, display: None } => writeln!(f, "=> {url}"),
            Self::Prompt {
                url,
                display: Some(display),
            } => writeln!(f, "=: {url} {display}"),
            Self::Prompt { url, display: None } => writeln!(f, "=: {url}"),
        }
    }
}

impl Node {
    pub fn write_html<W: Write>(&self, mut writer: W) -> fmt::Result {
        match self {
            Self::Text(s) => {
                writeln!(writer, "<p>")?;
                textwrap::wrap(s, 80)
                    .into_iter()
                    .try_for_each(|line| writeln!(writer, "{line}"))?;
                writeln!(writer, "</p>")
            }
            Self::H1(s) => writeln!(writer, "<h1>{s}</h1>"),
            Self::H2(s) => writeln!(writer, "<h2>{s}</h3>"),
            Self::H3(s) => writeln!(writer, "<h3>{s}</h3>"),
            Self::ListItem(s) => {
                write!(writer, "<li>")?;
                textwrap::wrap(s, 80)
                    .iter()
                    .try_for_each(|line| write!(writer, "{line}"))?;
                writeln!(writer, "</li>")
            }
            Self::Quote(s) => {
                write!(writer, "<blockquote>")?;
                textwrap::wrap(s, 80)
                    .iter()
                    .try_for_each(|line| writeln!(writer, "{line}"))?;
                writeln!(writer, "</blockquote>")
            }
            Self::PreStart { alt: Some(s) } => {
                writeln!(writer, "<pre title=\"{s}\">")
            }
            Self::PreStart { alt: None } => writeln!(writer, "<pre>"),
            Self::PreBody(s) => writeln!(writer, "{s}"),
            Self::PreEnd => writeln!(writer, "</pre>"),
            Self::Link { url, display } | Self::Prompt { url, display } => {
                writeln!(
                    writer,
                    "<a href=\"{url}\">{}</a>",
                    match display {
                        Some(d) => d,
                        None => url,
                    }
                )
            }
        }
    }
}
